@@ -5,27 +5,31 @@ Classes for drawing maps.
 
 import warnings
 from collections import OrderedDict
-from typing import Dict, List, Optional, Sequence, Tuple, Type, Union
+from typing import TYPE_CHECKING, Optional, Sequence, Union, cast
 
 from branca.element import Element, Figure, Html, MacroElement
-from jinja2 import Template
 
 from folium.elements import ElementAddToElement, EventHandler
+from folium.template import Template
 from folium.utilities import (
     JsCode,
     TypeBounds,
+    TypeBoundsReturn,
     TypeJsonValue,
-    camelize,
     escape_backticks,
     parse_options,
+    remove_empty,
     validate_location,
 )
 
+if TYPE_CHECKING:
+    from folium.features import CustomIcon, DivIcon
+
 
 class Evented(MacroElement):
     """The base class for Layer and Map
 
-    Adds the `on` method for event handling capabilities.
+    Adds the `on` and `once` methods for event handling capabilities.
 
     See https://leafletjs.com/reference.html#evented for
     more in depth documentation. Please note that we have
@@ -34,8 +38,14 @@ class Evented(MacroElement):
     """
 
     def on(self, **event_map: JsCode):
+        self._add(once=False, **event_map)
+
+    def once(self, **event_map: JsCode):
+        self._add(once=True, **event_map)
+
+    def _add(self, once: bool, **event_map: JsCode):
         for event_type, handler in event_map.items():
-            self.add_child(EventHandler(event_type, handler))
+            self.add_child(EventHandler(event_type, handler, once))
 
 
 class Layer(Evented):
@@ -109,7 +119,7 @@ class FeatureGroup(Layer):
         """
         {% macro script(this, kwargs) %}
             var {{ this.get_name() }} = L.featureGroup(
-                {{ this.options|tojson }}
+                {{ this.options|tojavascript }}
             );
         {% endmacro %}
         """
@@ -126,7 +136,7 @@ class FeatureGroup(Layer):
         super().__init__(name=name, overlay=overlay, control=control, show=show)
         self._name = "FeatureGroup"
         self.tile_name = name if name is not None else self.get_name()
-        self.options = parse_options(**kwargs)
+        self.options = remove_empty(**kwargs)
 
 
 class LayerControl(MacroElement):
@@ -180,7 +190,7 @@ class LayerControl(MacroElement):
             let {{ this.get_name() }} = L.control.layers(
                 {{ this.get_name() }}_layers.base_layers,
                 {{ this.get_name() }}_layers.overlays,
-                {{ this.options|tojson }}
+                {{ this.options|tojavascript }}
             ).addTo({{this._parent.get_name()}});
 
             {%- if this.draggable %}
@@ -201,7 +211,7 @@ class LayerControl(MacroElement):
     ):
         super().__init__()
         self._name = "LayerControl"
-        self.options = parse_options(
+        self.options = remove_empty(
             position=position, collapsed=collapsed, autoZIndex=autoZIndex, **kwargs
         )
         self.draggable = draggable
@@ -212,7 +222,7 @@ class LayerControl(MacroElement):
         self.base_layers = OrderedDict()
         self.overlays = OrderedDict()
 
-    def render(self, **kwargs) -> None:
+    def render(self, **kwargs):
         """Renders the HTML representation of the element."""
         self.reset()
         for item in self._parent._children.values():
@@ -263,7 +273,7 @@ class Icon(MacroElement):
         """
         {% macro script(this, kwargs) %}
             var {{ this.get_name() }} = L.AwesomeMarkers.icon(
-                {{ this.options|tojson }}
+                {{ this.options|tojavascript }}
             );
             {{ this._parent.get_name() }}.setIcon({{ this.get_name() }});
         {% endmacro %}
@@ -307,7 +317,7 @@ class Icon(MacroElement):
                 f"color argument of Icon should be one of: {self.color_options}.",
                 stacklevel=2,
             )
-        self.options = parse_options(
+        self.options = remove_empty(
             marker_color=color,
             icon_color=icon_color,
             icon=icon,
@@ -356,7 +366,7 @@ class Marker(MacroElement):
         {% macro script(this, kwargs) %}
             var {{ this.get_name() }} = L.marker(
                 {{ this.location|tojson }},
-                {{ this.options|tojson }}
+                {{ this.options|tojavascript }}
             ).addTo({{ this._parent.get_name() }});
         {% endmacro %}
         """
@@ -367,14 +377,14 @@ class Marker(MacroElement):
         location: Optional[Sequence[float]] = None,
         popup: Union["Popup", str, None] = None,
         tooltip: Union["Tooltip", str, None] = None,
-        icon: Optional[Icon] = None,
+        icon: Optional[Union[Icon, "CustomIcon", "DivIcon"]] = None,
         draggable: bool = False,
         **kwargs: TypeJsonValue,
     ):
         super().__init__()
         self._name = "Marker"
         self.location = validate_location(location) if location is not None else None
-        self.options = parse_options(
+        self.options = remove_empty(
             draggable=draggable or None, autoPan=draggable or None, **kwargs
         )
         if icon is not None:
@@ -387,15 +397,15 @@ class Marker(MacroElement):
                 tooltip if isinstance(tooltip, Tooltip) else Tooltip(str(tooltip))
             )
 
-    def _get_self_bounds(self) -> List[List[float]]:
+    def _get_self_bounds(self) -> TypeBoundsReturn:
         """Computes the bounds of the object itself.
 
         Because a marker has only single coordinates, we repeat them.
         """
         assert self.location is not None
-        return [self.location, self.location]
+        return cast(TypeBoundsReturn, [self.location, self.location])
 
-    def render(self) -> None:
+    def render(self):
         if self.location is None:
             raise ValueError(
                 f"{self._name} location must be assigned when added directly to map."
@@ -424,7 +434,7 @@ class Popup(Element):
 
     _template = Template(
         """
-        var {{this.get_name()}} = L.popup({{ this.options|tojson }});
+        var {{this.get_name()}} = L.popup({{ this.options|tojavascript }});
 
         {% for name, element in this.html._children.items() %}
             {% if this.lazy %}
@@ -476,14 +486,14 @@ class Popup(Element):
 
         self.show = show
         self.lazy = lazy
-        self.options = parse_options(
+        self.options = remove_empty(
             max_width=max_width,
             autoClose=False if show or sticky else None,
             closeOnClick=False if sticky else None,
             **kwargs,
         )
 
-    def render(self, **kwargs) -> None:
+    def render(self, **kwargs):
         """Renders the HTML representation of the element."""
         for name, child in self._children.items():
             child.render(**kwargs)
@@ -526,22 +536,11 @@ class Tooltip(MacroElement):
                 `<div{% if this.style %} style={{ this.style|tojson }}{% endif %}>
                      {{ this.text }}
                  </div>`,
-                {{ this.options|tojson }}
+                {{ this.options|tojavascript }}
             );
         {% endmacro %}
         """
     )
-    valid_options: Dict[str, Tuple[Type, ...]] = {
-        "pane": (str,),
-        "offset": (tuple,),
-        "direction": (str,),
-        "permanent": (bool,),
-        "sticky": (bool,),
-        "interactive": (bool,),
-        "opacity": (float, int),
-        "attribution": (str,),
-        "className": (str,),
-    }
 
     def __init__(
         self,
@@ -556,7 +555,7 @@ class Tooltip(MacroElement):
         self.text = str(text)
 
         kwargs.update({"sticky": sticky})
-        self.options = self.parse_options(kwargs)
+        self.options = remove_empty(**kwargs)
 
         if style:
             assert isinstance(
@@ -565,23 +564,6 @@ class Tooltip(MacroElement):
             # noqa outside of type checking.
             self.style = style
 
-    def parse_options(
-        self,
-        kwargs: Dict[str, TypeJsonValue],
-    ) -> Dict[str, TypeJsonValue]:
-        """Validate the provided kwargs and return options as json string."""
-        kwargs = {camelize(key): value for key, value in kwargs.items()}
-        for key in kwargs.keys():
-            assert (
-                key in self.valid_options
-            ), "The option {} is not in the available options: {}.".format(
-                key, ", ".join(self.valid_options)
-            )
-            assert isinstance(
-                kwargs[key], self.valid_options[key]
-            ), f"The option {key} must be one of the following types: {self.valid_options[key]}."
-        return kwargs
-
 
 class FitBounds(MacroElement):
     """Fit the map to contain a bounding box with the
@@ -660,7 +642,7 @@ class FitOverlays(MacroElement):
                 }
             });
             if (bounds.isValid()) {
-                {{ this._parent.get_name() }}.{{ this.method }}(bounds, {{ this.options|tojson }});
+                {{ this._parent.get_name() }}.{{ this.method }}(bounds, {{ this.options|tojavascript }});
             }
         }
         {{ this._parent.get_name() }}.on('overlayadd', customFlyToBounds);
@@ -682,7 +664,7 @@ class FitOverlays(MacroElement):
         self._name = "FitOverlays"
         self.method = "flyToBounds" if fly else "fitBounds"
         self.fit_on_map_load = fit_on_map_load
-        self.options = parse_options(padding=(padding, padding), max_zoom=max_zoom)
+        self.options = remove_empty(padding=(padding, padding), max_zoom=max_zoom)
 
 
 class CustomPane(MacroElement):
